// Copyright (c) 2013, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

part of parser;

/**
 * Print CSS as either pretty printed compact parameter = false (see visit's
 * compact optional parameter); compact is true then output is emitted for
 * efficient suite testing.
 */
class Emitter extends Visitor {
  final StringBuffer _buff = new StringBuffer();
  bool prettyPrint = true;

  void visit(var tree, [bool compact = true]) {
    prettyPrint = !compact;
    _buff.clear();
    visitStylesheet(tree);
  }

  /** Items to emit to the output buffer. */
  void emit(String str) {
    _buff.add(str);
  }

  String toString() => _buff.toString().trim();

  String get _newLine => prettyPrint ? '\n' : ' ';
  String get _sp => prettyPrint ? ' ' : '';

  // TODO(terry): When adding obfuscation we'll need isOptimized (compact w/
  //              obufuscation) and have isTesting (compact no obfuscation) and
  //              isCompact would be !prettyPrint.  We'll need another boolean
  //              flag for obfuscation.
  bool get _isTesting => !prettyPrint;

  void visitStylesheet(Stylesheet node) {
    for (final topLevel in node.topLevels) {
      topLevel.visit(this);
    }
  }

  void visitCssComment(CssComment node) {
    emit('/* ${node.comment} */');
  }

  void visitCommentDefinition(CommentDefinition node) {
    emit('<!-- ${node.comment} -->');
  }

  void visitMediaDirective(MediaDirective node) {
    emit(' @media');
    int idx = 0;
    for (var medium in node.media) {
      emit(idx++ == 0 ? ' ' : ',');
      medium.visit(this);
    }
    emit(' {');
    node.ruleset.visit(this);
    emit('$_newLine\}');
  }

  // @page : pseudoPage {
  //    decls
  // }
  void visitPageDirective(PageDirective node) {
    emit('$_newLine@page');
    if (node.hasIdent || node.hasPseudoPage) {
      if (node.hasIdent) emit(' ');
      emit(node._ident);
      emit(node.hasPseudoPage ? ':${node._pseudoPage}' : '');
    }
    emit(' ');
    int idx = 0;
    for (var decl in node._declsMargin) {
      if (idx > 0) emit(_newLine);
      emit('{$_newLine');
      decl.visit(this);
      emit('}');
      idx++;
    }
  }

  void visitImportDirective(ImportDirective node) {
    bool isStartingQuote(String ch) => ('\'"'.indexOf(ch) >= 0);

    if (isStartingQuote(node._import)) {
      emit(' @import "${node._import}"');
    } else {
      if (_isTesting) {
        // Emit exactly was we parsed.
        emit(' @import url(${node._import})');
      } else {
        // url(...) isn't needed only a URI can follow an @import directive.
        emit(' @import ${node._import}');
      }
    }

    int idx = 0;
    for (final medium in node._media) {
      emit(idx++ == 0 ? '$_sp' : ',');
      medium.visit(this);
    }
    emit(';');
  }

  void visitKeyFrameDirective(KeyFrameDirective node) {
    emit('$_newLine@-webkit-keyframes ');
    node._name.visit(this);
    emit('$_sp{$_newLine');
    for (final block in node._blocks) {
      block.visit(this);
    }
    emit('}');
  }

  void visitKeyFrameBlock(KeyFrameBlock node) {
    emit('$_sp$_sp');
    node._blockSelectors.visit(this);
    emit('$_sp{$_newLine');
    node._declarations.visit(this);
    emit('$_sp$_sp}$_newLine');
  }

  void visitFontFaceDirective(FontFaceDirective node) {
    // TODO(terry): To Be Implemented
  }

  void visitIncludeDirective(IncludeDirective node) {
    emit('/****** @include ${node._include} ******/\n');
    if (node._stylesheet != null) {
      node._stylesheet.visit(this);
    } else {
      emit('// <EMPTY>');
    }
    emit('/****** End of ${node._include} ******/\n\n');
  }

  void visitStyletDirective(StyletDirective node) {
    emit('/* @stylet export as ${node._dartClassName} */\n');
  }

  void visitNamespaceDirective(NamespaceDirective node) {
    bool isStartingQuote(String ch) => ('\'"'.indexOf(ch) >= 0);

    if (isStartingQuote(node._uri)) {
      emit(' @namespace ${node.prefix}"${node._uri}"');
    } else {
      if (_isTesting) {
        // Emit exactly was we parsed.
        emit(' @namespace ${node.prefix}url(${node._uri})');
      } else {
        // url(...) isn't needed only a URI can follow a:
        //    @namespace prefix directive.
        emit(' @namespace ${node.prefix}${node._uri}');
      }
    }
    emit(';');
  }

  void visitRuleSet(RuleSet node) {
    emit("$_newLine");
    node._selectorGroup.visit(this);
    emit(" {$_newLine");
    node._declarationGroup.visit(this);
    emit("}");
  }

  void visitDeclarationGroup(DeclarationGroup node) {
    int idx = 0;
    for (final declaration in node._declarations) {
      if (idx > 0) emit(_newLine);
      emit("$_sp$_sp");
      declaration.visit(this);
      emit(";");
      idx++;
    }
    if (idx > 0) emit(_newLine);
  }

  void visitMarginGroup(MarginGroup node) {
    var margin_sym_name =
        TokenKind.idToValue(TokenKind._MARGIN_DIRECTIVES, node.margin_sym);

    emit("@$margin_sym_name {$_newLine");

    visitDeclarationGroup(node);

    emit("}$_newLine");
  }

  void visitDeclaration(Declaration node) {
    String importantAsString() => node.important ? '$_sp!important' : '';

    emit("${node._property.name}: ");
    node._expression.visit(this);

    emit("${importantAsString()}");
  }

  void visitSelectorGroup(SelectorGroup node) {
    int idx = 0;
    for (final selector in node._selectors) {
      if (idx++ > 0) {
        emit(',$_sp');
      }
      selector.visit(this);
    }
  }

  void visitSimpleSelectorSequence(SimpleSelectorSequence node) {
    emit('${node._combinatorToString}');
    node._selector.visit(this);
  }

  void visitSimpleSelector(SimpleSelector node) {
    emit(node.name);
  }

  void visitNamespaceSelector(NamespaceSelector node) {
    emit("${node.namespace}|${node.nameAsSimpleSelector.name}");
  }

  void visitElementSelector(ElementSelector node) {
    emit("${node.name}");
  }

  void visitAttributeSelector(AttributeSelector node) {
    emit("[${node.name}${node.matchOperator()}${node.valueToString()}]");
  }

  void visitIdSelector(IdSelector node) {
    emit("#${node.name}");
  }

  void visitClassSelector(ClassSelector node) {
    emit(".${node.name}");
  }

  void visitPseudoClassSelector(PseudoClassSelector node) {
    emit(":${node.name}");
  }

  void visitPseudoElementSelector(PseudoElementSelector node) {
    emit("::${node.name}");
  }

  void visitNotSelector(NotSelector node) {
    // TODO(terry): TBD
  }

  void visitLiteralTerm(LiteralTerm node) {
    emit(node._text);
  }

  void visitHexColorTerm(HexColorTerm node) {
    var mappedName;
    if (_isTesting && !(node.value is BAD_HEX_VALUE)) {
      mappedName = TokenKind.hexToColorName(node.value);
    }
    if (mappedName == null) {
      mappedName = '#${node.text}';
    }

    emit(mappedName);
  }

  void visitNumberTerm(NumberTerm node) {
    visitLiteralTerm(node);
  }

  void visitUnitTerm(UnitTerm node) {
    emit('${node.text}${node.unitToString()}');
  }

  void visitLengthTerm(LengthTerm node) {
    emit('${node.text}${node.unitToString()}');
  }

  void visitPercentageTerm(PercentageTerm node) {
    emit('${node.text}%');
  }

  void visitEmTerm(EmTerm node) {
    emit('${node.text}em');
  }

  void visitExTerm(ExTerm node) {
    emit('${node.text}ex');
  }

  void visitAngleTerm(AngleTerm node) { }

  void visitTimeTerm(TimeTerm node) { }

  void visitFreqTerm(FreqTerm node) { }

  void visitFractionTerm(FractionTerm node) {
    emit('${node.text}fr');
  }

  void visitUriTerm(UriTerm node) {
    emit('url("${node.text}")');
  }

  void visitFunctionTerm(FunctionTerm node) {
    // TODO(terry): Optimize rgb to a hexcolor.
    emit('${node.text}(');
    node._params.visit(this);
    emit(')');
  }

  void visitGroupTerm(GroupTerm node) {
    emit('(');
    int idx = 0;
    for (final term in node._terms) {
      if (idx++ > 0) {
        emit('$_sp');
      }
      term.visit(this);
    }
    emit(')');
  }

  void visitItemTerm(ItemTerm node) {
    emit('[${node.text}]');
  }

  void visitOperatorSlash(OperatorSlash node) {
    emit('/');
  }

  void visitOperatorComma(OperatorComma node) {
    emit(',');
  }

  void visitExpression(Expression node) {

  }

  void visitExpressions(Expressions node) {
    int idx = 0;
    for (final expression in node._expressions) {
      // Add space seperator between terms without an operator.
      // TODO(terry): Should have a BinaryExpression to solve this problem.
      if (idx > 0 &&
          !(expression is OperatorComma || expression is OperatorSlash)) {
        emit(' ');
      }
      expression.visit(this);
      idx++;
    }
  }

  void visitBinaryExpression(BinaryExpression node) { }

  void visitUnaryExpression(UnaryExpression node) { }

  void visitIdentifier(Identifier node) {
    emit(node.name);
  }

  void visitWildcard(Wildcard node) {
    emit('*');
  }

  // TODO(terry): Defined for frog/tree.dart.
  void visitTypeReference(TypeReference node) { }

  void visitDartStyleExpression(DartStyleExpression node) { }

  void visitFontExpression(FontExpression node) { }

  void visitBoxExpression(BoxExpression node) { }

  void visitMarginExpression(MarginExpression node) { }

  void visitBorderExpression(BorderExpression node) { }

  void visitHeightExpression(HeightExpression node) { }

  void visitPaddingExpression(PaddingExpression node) { }

  void visitWidthExpression(WidthExpression node) { }
}
