// Copyright (c) 2013, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

library visitor_test;

import 'dart:utf';
import 'package:unittest/unittest.dart';
import 'package:unittest/vm_config.dart';
import 'package:csslib/parser.dart';
import 'testing.dart';


// Pretty printer for CSS.
var emitCss = new Emitter();
String prettyPrint(Stylesheet ss) => (emitCss..visit(ss, false)).toString();

class ClassVisitor extends Visitor {
  final List expectedClasses;
  final List foundClasses = [];

  ClassVisitor(this.expectedClasses);

  void visit(var tree) => super.visitStylesheet(tree);

  void visitClassSelector(ClassSelector node) {
    if (!foundClasses.contains(node.name)) {
      foundClasses.add(node.name);
    }
  }

  bool get matches => expectedClasses.toString() == foundClasses.toString();
}

void testClassVisitors() {
  var cssErrors = [];

  var in1 = '.foobar { }';

  var s = parseCss(in1, errors: cssErrors);

  var clsVisits = new ClassVisitor(['foobar'])..visit(s);
  expect(clsVisits.matches, true);

  cssErrors = [];
  in1= '''
      .foobar1 { }
      .xyzzy .foo #my-div { color: red; }
      div.hello { font: arial; }
    ''';

  s = parseCss(in1, errors: cssErrors);
  clsVisits = new ClassVisitor(['foobar1', 'xyzzy', 'foo', 'hello'])..visit(s);
  expect(clsVisits.matches, true);

  expect(prettyPrint(s), r'''
.foobar1 {
}
.xyzzy .foo #my-div {
  color: #ff0000;
}
div.hello {
  font: arial;
}''');
}

class PolyfillEmitter extends Emitter {
  final String _prefix;

  PolyfillEmitter(this._prefix);

  void visit(var tree, [bool compact = true]) => super.visitStylesheet(tree);

  void visitClassSelector(ClassSelector node) {
    emit('.${_prefix}_${node.name}');
  }
}

String polyfillPrint(String prefix, Stylesheet ss) =>
  (new PolyfillEmitter(prefix)..visit(ss, false)).toString();

void testPolyFill() {
  var cssErrors = [];

  final input = r'''
.foobar { }
div.xyzzy { }
#foo .foo .bar .foobar { }
''';

  final generated = r'''
.myComponent_foobar {
}
div.myComponent_xyzzy {
}
#foo .myComponent_foo .myComponent_bar .myComponent_foobar {
}''';

  var s = parseCss(input, errors: cssErrors);

  final emitted = polyfillPrint('myComponent', s);
  expect(emitted, generated);
}

main() {
  useVmConfiguration();
  useMockMessages();

  test('Class Visitors', testClassVisitors);
  test('Polyfill', testPolyFill);
}
